//-----------------------------------------------------------------------------
/// @file rf_expr.hh
///
/// Refal+ expression interface
//
//  $Source$
//  $Revision: 2187 $
//  $Date: 2006-12-15 03:17:56 +0300 (Fri, 15 Dec 2006) $
//-----------------------------------------------------------------------------

#ifndef __rf_expr_hh__
#define __rf_expr_hh__

#include "rf_common.hh"
#include "rf_term.hh"
#include "rf_memory_chunk.hh"
#include "rf_object.hh"

#include <new>
#include <string.h>


///
/// A namespace for Refal runtime
namespace rfrt
{

class Result;

class True;

using namespace rftype ;

using pxx::WString;

///
/// "Flat" bit. This bit, if set for reference term, means that referenced
/// expression is "flat", i.e. it contains only short symbols, so we should
/// accurately handle this when destroying the expression. This bit was
/// introduced because in a case of a "flat" expression we don't need to walk
/// through and dereference childs.
/// This bit is kept as invariant during all expression operations.
#define FLAT_BIT (uintptr_t(0x04))

///
/// Parameter enumeration for SplitIterator template class
enum Direction {
  d_lt,         ///< Split from left
  d_rt          ///< Split from right
};

///
/// Refal expression class. This is the main type of Refal runtime objects,
/// and the only one the user should deal with. Other object types are not
/// visible from generated code and shouldn't be used directly.
class Expr
{

///
/// We may use expression methods in SplitIterator class. We cannot avoid this
/// because SplitIterator simultaneously changes two expressions in a way that
/// adding necessary methods to Expr class will break invariants.
template <Direction D = d_lt>
friend class SplitIterator ;
///
/// Stack and Arg classes are our friends because they use special copy
/// constructor which shouldn't be public available.
friend class Arg ;
friend class Stack ;
friend class Term ;

private:
  ///
  /// Pointer to the first term
  Term* first;
  ///
  /// Pointer beyond the last term
  Term* last;
  ///
  /// Pointer to a memory block containing an expression
  MemoryChunk* mem_chunk;
  ///
  /// Various expression flags. Currently only FLAT_BIT is supported, meaning
  /// that expression is "flat", i.e. it can't contain reference terms.
  uintptr_t flags;

public:

#if 0
  ///
  /// Constructor from character string
  inline Expr (char const* _string = null) ;
#endif
  ///
  /// Create an empty expression
  inline Expr () ;
  ///
  /// Constructor from expression holder. This constructor does dereference
  /// operation.
  /// @param _cp        Pointer to a term containg a reference to expression
  ///                   to be dereferenced
  inline Expr (Term const* _cp) ;
  ///
  /// Dereference constructor from explicit term in expression. Should be
  /// called only with valid index argument.
  /// @param _expr      Source expression
  /// @param _index     Term index inside expression
  inline Expr (Expr const& _expr, uintptr_t _index) ;
  ///
  /// Subexpression constructor
  /// @param _expr      Source expression
  /// @param _index     Index where subexpression started
  /// @param _length    Subexpression length
  inline Expr (Expr const& _expr, uintptr_t _index, uintptr_t _length) ;
  ///
  /// Copy constructor
  /// @param _expr      Source expression
  inline Expr (Expr const& _expr) ;
  ///
  /// Constructor from object
  /// @param _obj       Source object
  inline Expr (Object* _obj) ;
  ///
  /// Destructor
  inline ~Expr () ;
  ///
  /// Assignment operator
  /// @param _expr      Right side expression
  inline Expr& operator = (Expr const& _expr) ;
  ///
  /// Expression concatenation
  INLINE Expr operator + (Expr const& _expr) const ;
  ///
  /// Operator enclosing an expression into parentheses
  inline Expr operator () () const ;
  ///
  /// Operator that dereferences expression containing a single reference term
  inline Expr operator * () const ;
  ///
  /// Check whether a term at specified index is a symbol
  inline bool symbol_at (uintptr_t _index) const ;
  ///
  /// Check whether an expression is flat
  inline bool is_flat () const ;
  ///
  /// Check whether an expression is empty
  inline bool is_empty () const ;
  ///
  /// Get a pointer to the first term
  inline Term* get_first () const ;
  ///
  /// Get a pointer beyond the last term
  inline Term* get_last () const ;
  ///
  /// Get expression length (in terms)
  inline uintptr_t get_len () const ;
private:
  ///
  /// Get expression flags
  inline uintptr_t get_flags () const ;
public:
  ///
  /// Dump expression contents (not recursive)
  inline void dump () const ;
  ///
  /// Write an expression to buffered file stream. This function is incomplete
  /// and suboptimal.
  INLINE bool write (FILE* _fp) const ;
  ///
  /// The same as write(), but adds new line at the end of output.
  inline bool writeln (FILE* _fp) const ;
  INLINE bool print (FILE* _fp) const ;
  inline bool println (FILE* _fp) const ;
  INLINE operator WString () const ;
  ///
  /// Compute expression hash
  inline uint32_t hash () const ;
  ///
  /// Compare with subexpression of other expression.  Attention: doesn't check
  /// that our expression is not longer then the rest of the other.
  inline bool eq (Expr const& _expr, uintptr_t _index) ;
  ///
  /// Same method for const expression.
  inline bool eq (Expr const& _expr, uintptr_t _index) const ;
  ///
  /// Compare our first term with some one of another expression.
  /// Should be called only with valid index argument.
  inline bool term_eq (Expr const& _expr, uintptr_t _index) const ;
  ///
  /// Equality check operator
  inline bool operator == (Expr& _expr) ;
  ///
  /// Inequality check operator
  inline bool operator != (Expr& _expr) ;
  ///
  /// Same methods for const expressions.
  inline bool operator == (Expr const& _expr) const ;
  inline bool operator != (Expr const& _expr) const ;
  ///
  /// Compare two expressions.  Return integer > 0 if first expression is
  /// larger then second one, 0 if they are equal, and integer < 0 if first one
  /// is smaller then second one.
  static inline int compare (Expr const& _expr1, Expr const& _expr2) ;
  ///
  /// Clear memory occupied by the pointed expression.
  /// This method is used to clear a space for uninitialized results on a
  /// stack, so if a refal function returns "$fail" we can distinguish
  /// whether a result expression was already constructed and delete it.
  static inline void clear (Expr* _ptr) ;
  ///
  /// Try to get a reference to the object underlying the first term.
  /// If the term has a type other then RF_SYMBOL(ObjClass) then throw
  /// InvalidParam().
  template <class ObjClass>
  inline ObjClass const& get_object () const ;
  ///
  /// Try to cast the first term to the concrete term type
  /// (RF_SYMBOL(ObjClass)) and return a pointer to the underlying object.
  /// If not success return null.
  template <class ObjClass>
  inline ObjClass* try_get_ptr () const ;
  ///
  /// If a type of the first term is RF_SYMBOL(ObjClass1) then convert the
  /// underlying object to ObjClass2.
  /// Else throw InvalidParam().
  template <class ObjClass1, class ObjClass2>
  inline ObjClass2 convert_object() const ;
  ///
  /// Same as convert_object() but before converting check that Cond::check()
  /// is true for the object.
  template <class ObjClass1, class ObjClass2, class Cond>
  inline ObjClass2 check_convert_object() const ;

  template<class TermClass, typename CppType>
  static inline Expr create_seq (CppType const* _array, size_t _size) ;

  template<class TermClass, typename CppType>
  static inline Expr create_seq (CppType const* _array) ;

  template<class ObjClass, typename CppType1>
  static inline Expr create_terms (CppType1 const* _array) ;

  template<class ObjClass, typename CppType1, typename CppType2>
  static inline Expr create_terms (CppType1 const* _array, CppType2 const& _arg2) ;

  template<class ObjClass>
  static inline Expr create () ;

  template<class ObjClass1, class ObjClass2>
  inline Expr convert () const ;

  template<class ObjClass1, class ObjClass2>
  inline Expr (Expr const& _expr) ;

  template<class ObjClass, class CppType1>
  static inline Expr create (const CppType1& _arg1) ;

  template<class ObjClass, class CppType1, class CppType2>
  static inline Expr create (const CppType1& _arg1, const CppType2& _arg2) ;

  template <term_class_t, class ObjClass>
  friend class CreateExpr0;

  template<term_class_t, class ObjClass, class CppType1>
  friend class CreateExpr1;

  template<term_class_t, class ObjClass, class CppType1, class CppType2>
  friend class CreateExpr2;

  template <class ObjClass>
  static inline Expr create_sym () ;

  template <class ObjClass, class CppType1>
  static inline Expr create_sym (const CppType1& _arg1) ;

  template <class ObjClass, class CppType1, class CppType2>
  static inline Expr create_sym (const CppType1& _arg1, const CppType2& _arg2) ;
  
  template <class ObjClass, class CppType1, class CppType2, class CppType3>
  static inline Expr create_sym (
      const CppType1& _arg1, const CppType2& _arg2, const CppType3& _arg3) ;
  
  template <class ObjClass>
  static inline Expr create_static_sym (wchar_t const* _name) ;

  template <class TypeCond, class ObjCond>
  inline bool check () const ;

  template <class TypeCond>
  inline bool check_type () const ;

private:

  template <class T>
  inline void set_flat_bit () ;

private:

  ///
  /// Create an unitialized expression with given length. This constructor is
  /// private because expression body contains garbage. It is used during
  /// concatenation to allocate memory where resulting expression will be
  /// constructed.
  /// @param _len       A length of expression
  /// @param _align     Indicates how an expression will be aligned in a memory
  ///                   block: < 0 for left alignment, > 0 for right alignment,
  ///                   0 for centered alignment
  inline Expr (size_t _len, int _align = 0) ;
  ///
  /// Generic constructor. Should not be called directly because its arguments
  /// may not be arbitrary.
  /// @param _first     Pointer to the first term
  /// @param _last      Pointer to the right margin
  /// @param _mem_chunk Pointer to the memory block containing an expression
  /// @param _flags     Expression flags
  inline Expr (
    Term* const _first, Term* const _last,
    MemoryChunk* const _mem_chunk, uintptr_t const _flags
  ) ;
  ///
  /// Constructor from a reference term. This constructor originally was
  /// public because operator () returned Term instead of Expr, so this
  /// constructor was actually called from user code. Now the situation is
  /// changed and operator () now returns Expr. This allows to avoid friend
  /// operators + () with Term arguments. For a while we left that code,
  /// because further we may want to use it for optimization. Though we believe
  /// that compiler should optimize a creation of temporary objects.
  /// @param _cp        A term to become an expression.
  inline Expr (Term const& _cp) ;
  ///
  /// A special copy constructor from expression pointer. This constructor is
  /// used only to pop arguments and results from a stack. The trick is that
  /// we don't increment memory block reference counter when popping arguments
  /// from a stack. This allows to simply shift a stack pointer rather then
  /// doing full stack unwind.
  /// @param _expr      Source expression
  inline Expr (Expr const* _expr) ;
  ///
  /// Initialize new expression with specified length and alignment
  /// @param _len       A length of expression
  /// @param _align     Indicates how an expression will be aligned in a memory
  ///                   block: < 0 for left alignment, > 0 for right alignment,
  ///                   0 for centered alignment
  INLINE void init (size_t _len, int _align) ;
  ///
  /// Drop an expression contents. Used in destructor and assignment operator.
  /// After a call to this method expression may be used only to create a new
  /// expression at old place.
  inline void drop () ;
  ///
  /// Increment reference counters for all expressions referenced from us
  inline void ref_childs () const ;
  ///
  /// Decrement reference counters for all expressions referenced from us.
  /// This function is called when we lost the last reference to our memory
  /// block, so it should not only walk through our expression but also look
  /// beside expression margins.
  INLINE void deref_childs () const ;
  static INLINE void deref_childs (
    Term* _first, Term* _last, MemoryChunk* _mem
  ) ;
  ///
  /// Get pointer to a memory block containing an expression
  inline MemoryChunk* get_mem_chunk () const ;
  ///
  /// Check whether we can copy another expression to the right from our
  /// in the same memory block.
  inline bool rt_alloc (uintptr_t _len) const ;
  ///
  /// Check whether we can copy another expression to the left from our
  /// in the same memory block.
  inline bool lt_alloc (uintptr_t _len) const ;

protected:

  ///
  /// Set term at _p as a border.  To use in case of expression construction
  /// failure.
  static inline void set_border (Term* _p) ;

public:

  class TooLargeArgument :
    public pxx::Exception
  {
  public:
    void print()
      { fprintf(stderr, "Argument too large for conversion\n"); }
  };

  class InvalidParam :
    public pxx::Exception
  {
  public:
    void print()
      { fprintf(stderr, "Invalid parameter\n"); }
  };

  template <class C>
  class Iterator
  {
    Term* curr;
    Term* last;
  public:
    Iterator (Expr const& _expr) :
      curr (_expr.get_first()),
      last (_expr.get_last())
    {}
    ///
    inline Iterator& operator ++ ()
    {
      curr++;
      return self;
    }
    ///
    inline Iterator& operator ++ (int)
    {
      return operator++();
    }
    ///
    inline operator bool () const
    {
      return curr < last;
    }
  };

  template <class TypeCond, class Conv, class Cond = True>
  class Convert;

  inline Expr const& operator , (Expr const& _expr) const ;
  inline Expr const& operator , (Expr const& _expr) ;
  inline Expr& operator , (Expr& _expr) ;
  inline Result& operator , (Result& _res) ;

};


struct ExprWrapper
{
  uintptr_t ptr;
  Expr      expr;
};

extern pxx::ChunkAllocator<ExprWrapper>* expr_allocator;

///
/// A class representing a splitting of expression into two parts. This class
/// is needed to work with "open" expressions. It is closely related to
/// expression itself, so it is contained in the same file.
template <Direction D>
class SplitIterator
{
  ///
  /// Left side of splitted expression
  Expr lt ;
  ///
  /// Right side of splitted expression
  Expr rt ;
  ///
  /// Iterator state
  bool state ;
  ///
  /// Copy constructor and assignment operator do not have any sense for this
  /// class.
  NO_COPY_CTOR(SplitIterator)
  NO_ASSIGN(SplitIterator)

public:
  ///
  /// Constructor
  inline SplitIterator (Expr const& _expr, uintptr_t _min_len = 0) ;
  ///
  /// Destructor
  ~SplitIterator () ;
  ///
  /// Increment iterator
  inline SplitIterator& operator ++ () ;
  ///
  /// Postfix form of increment iterator
  inline SplitIterator& operator ++ (int) ;
  ///
  /// Decrement iterator
  inline SplitIterator& operator -- () ;
  ///
  /// Postfix form of decrement iterator
  inline SplitIterator& operator -- (int) ;
  ///
  /// Get iterator state
  inline operator bool () ;
  ///
  /// Get left part of expression
  inline Expr const& get_left () const ;
  ///
  /// Get right part of expression
  inline Expr const& get_right () const ;

};

///
/// A macro naming iterator for expression variable
#define RF_iter(e) e##_iter
///
/// A macro splitting an expression variable
#define RF_lsplit(e, minlen, le, re) \
  SplitIterator<d_lt> RF_iter(e)(e, minlen); \
  Expr const& le = RF_iter(e).get_left(); \
  Expr const& re = RF_iter(e).get_right()
  
#define RF_rsplit(e, minlen, le, re) \
  SplitIterator<d_rt> RF_iter(e)(e, minlen); \
  Expr const& le = RF_iter(e).get_left(); \
  Expr const& re = RF_iter(e).get_right()


extern const Expr empty;
extern Expr unexpected_fail;
extern Expr trap_stack;

}


#endif // __rf_expr_hh__
